# frozen_string_literal: true

# Cookbook:: gitlab-ci-runner
# Library:: api
#
# Copyright:: 2018, Ben Hughes <bmhughes@bmhughes.co.uk>, All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'uri'
require 'net/http'

require_relative '_common'

module GitlabCiRunner
  # Helper module for polling the Gitlab API to configure runners
  module ApiHelpers
    include GitlabCiRunner::CommonHelpers

    API_BASE_URL = 'api/v4'
    RUNNER = "#{API_BASE_URL}/runners"
    RUNNER_ALL = "#{RUNNER}/all"
    RUNNER_VERIFY = "#{RUNNER}/verify"

    RUNNER_REGISTER_OPTIONS = %w(
      token
      description
      info
      active
      locked
      run_untagged
      tag_list
      access_level
      maximum_timeout
    ).freeze

    RUNNER_UPDATE_OPTIONS = %w(
      description
      active
      tag_list
      run_untagged
      locked
      access_level
      maximum_timeout
    ).freeze

    def api_runner_get_all(host, token)
      url = "#{host}/#{RUNNER_ALL}"

      api_request(type: :get, url: url, token: token)
    end

    def api_runner_get(host, token, id)
      url = "#{host}/#{RUNNER}/#{id}"

      api_request(type: :get, url: url, token: token)
    end

    def api_runner_valid?(host, token)
      url = "#{host}/#{RUNNER_VERIFY}"

      api_response_is?(
        api_request(
          type: :post,
          url: url,
          data: { 'token' => token }
        ),
        Net::HTTPSuccess
      )
    end

    def api_runner_register(host, options)
      invalid_options = options.keys.difference(RUNNER_REGISTER_OPTIONS)

      if invalid_options.empty?
        api_request(type: :post, url: "#{host}/#{RUNNER}", data: options)
      else
        raise_error(
          ArgumentError,
          'Invalid runner registration options '\
          "passed: #{invalid_options.join(', ')}"
        )
      end
    end

    def api_runner_update!(host, token, id, options)
      invalid_options = options.keys.difference(RUNNER_UPDATE_OPTIONS)

      if invalid_options.empty?
        api_request(
          type: :put,
          url: "#{host}/#{RUNNER}/#{id}",
          token: token,
          data: options
        )
      else
        raise_error(
          ArgumentError,
          "Invalid runner update options passed: #{invalid_options.join(', ')}"
        )
      end
    end

    def api_runner_remove!(host, token)
      url = "#{host}/#{RUNNER}"

      api_request(type: :delete, url: url, data: { 'token' => token })
    end

    def api_response_is?(response, type)
      if response.is_a?(type)
        log_chef(
          :debug,
          "Response #{response} #{response.code} "\
          "matches expected type #{type}."
        )
        return true
      end

      log_chef(
        :error,
        "Response #{response} #{response.code} error "\
        ", expected type #{type}."
      )
      false
    end

    private

    API_REQUEST_TYPES = %i(get put post delete).freeze
    API_REQUEST_TYPES_DATA = %i(put post delete).freeze
    API_USER_AGENT = 'gitlab-ci-runner Chef Cookbook'

    def http_create(url)
      uri = URI.parse(url)
      http = Net::HTTP.new(uri.host, uri.port)

      if uri.is_a?(URI::HTTPS)
        log_chef(:info, "Using TLS for #{uri.host}:#{uri.port}.")
        http.use_ssl = true
      end

      http
    end

    def request_create(type, url)
      Object.const_get(
        "Net::HTTP::#{type.to_s.capitalize}"
      ).send(:new, URI.parse(url).request_uri)
    end

    def request_set_headers(request, token)
      request['User-Agent'] = API_USER_AGENT
      request['Accept'] = 'application/json'
      request['Private-Token'] = token unless token.nil?

      request.each_header do |header, value|
        log_chef(:debug, "Header: #{header}: #{value}.")
      end

      request
    end

    def api_request(type:, url:, token: nil, data: nil)
      if API_REQUEST_TYPES.include?(type)
        log_chef(
          :info,
          "Performing API #{type} request to #{url} with "\
          "token: '#{token}' and data: '#{data}'."
        )

        request = request_create(type, url)
        request.set_form_data(data) if API_REQUEST_TYPES_DATA.include?(type)
        request = request_set_headers(request, token)

        response = http_create(url).request(request)
        log_chef(:debug, "Response code #{response.code}")

        response
      else
        raise_error(
          type: ArgumentError,
          message: "Invalid API request type #{type}"
        )
      end
    end
  end
end
