# frozen_string_literal: true

# Cookbook:: gitlab-ci-runner
# Library:: runner
#
# Copyright:: 2018, Ben Hughes <bmhughes@bmhughes.co.uk>, All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'toml'
require 'json'

require_relative 'api'
require_relative '_common'
require_relative 'toml'

module GitlabCiRunner
  # Helper methods for configuring gitlab-runner via chef
  module RunnerHelpers
    include GitlabCiRunner::ApiHelpers
    include GitlabCiRunner::CommonHelpers
    include GitlabCiRunner::TomlHelpers

    RUNNER_REGEXP =
      /(docker|parallels|virtualbox|ssh|machine|kubernetes)(_|-)/.freeze

    RUNNER_SERVER_OPTIONS =
      %w(
        description
        active
        tag_list
        run_untagged
        locked
        access_level
        maximum_timeout
      ).freeze

    RUNNER_LOCAL_OPTIONS =
      %w(
        name
        url
      ).freeze

    RUNNER_LOCAL_REMOVE =
      %w(
        registration_token
      ).freeze

    def feed_options(options_hash, prefix)
      options = []
      options_hash.each do |key, value|
        next if value.is_a?(String) && value.empty?

        if value.is_a?(Hash)
          options.concat(feed_options(value, format_keyname(key.to_s, prefix)))
        else
          options.concat(transform_option(format_keyname(key, prefix), value))
        end
      end
      options
    end

    def transform_option(key, value)
      if value.is_a?(TrueClass) || value.is_a?(FalseClass)
        ["--#{key}=#{value}"]
      elsif key.eql?('tag-list')
        ["--#{key} '#{value.join(',')}'"]
      elsif value.is_a?(Array)
        value.map { |item| "--#{key} '#{item}'" }
      else
        ["--#{key} '#{value}'"]
      end
    end

    def runner_update_local(config, options)
      unless parameter_defined?(config)
        log_chef(:debug, 'No runners configured.')
        return config
      end

      log_chef(
        :info,
        "Checking options for runner #{config['name']}."
      )
      options.each do |option, value|
        next unless parameter_defined?(value)

        log_chef(:debug, "Testing option #{option}.")
        if config[option].eql?(value)
          log_chef(
            :debug,
            "Option #{option} matches. "\
            "Current: #{config[option]}. Incoming: #{value})."
          )
          next
        end

        log_chef(
          :info,
          "Option #{option} has changed. "\
          "Current: #{config[option]}. Incoming: #{value}."
        )
        config[option] = value
      end

      option_cleanup(config: config, options: options)
    end

    def runner_options_to_h(options)
      options.keys.each do |option|
        next unless runner?(option)

        a = runner_option_split(option)
        options[a[0]] ||= {}
        options[a[0]][a[1]] = options.delete(option)
      end

      options
    end

    def runner_registered?(name, host, config_file)
      return false unless runner_configured?(name, host, config_file)

      token, url = runner_get_token(name, config_file)

      log_chef(:info, "Checking if runner #{name} is registered with #{host}.")
      if api_runner_valid?(url, token)
        log_chef(:info, "Runner #{name} is registered with #{host}.")
        true
      else
        log_chef(:info, "Runner #{name} is NOT registered with #{host}.")
        false
      end
    end

    def runner_configured?(name, host, config_file)
      token, url = runner_get_token(name, config_file)

      if !token.nil? && url.eql?(host)
        log_chef(
          :info,
          "Runner #{name} for host #{host} is configured in #{config_file}."
        )
        return true
      end

      log_chef(
        :info,
        "Runner #{name} for host #{host} is NOT configured in #{config_file}."
      )

      false
    end

    def runner_register(host, registration_token, options)
      if parameter_defined?(registration_token)
        options['token'] = registration_token

        response = api_runner_register(host, option_join(options))

        if api_response_is?(response, Net::HTTPCreated)
          response = JSON.parse(response.body)
          log_chef(
            :info,
            "Runner #{response['id']} created sucessfully on #{host}."
          )

          [response['id'], response['token']]
        else
          raise_error(
            message: "#{response.code} Error when registering: #{response}"
          )
        end
      else
        raise_error(
          type: ArgumentError,
          message: 'Registration requires a registration token to be passed'
        )
      end
    end

    def runner_unregister!(host, token)
      response = api_runner_remove!(host, token)

      if api_response_is?(response, Net::HTTPNoContent)
        log_chef(
          :info,
          "Runner with token #{token} successfully unregistered from #{host}."
        )
        return true
      end

      log_chef(
        :warn,
        "Error unregistering runner with token #{token} from #{host}."
      )

      false
    end

    def runner_update_server(host, token, id, options)
      response = api_runner_update!(host, token, id, option_join(options))

      if api_response_is?(response, Net::HTTPSuccess)
        log_chef(
          :info,
          "Runner #{id} updated successfully on #{host}."
        )
        return true
      end

      log_chef(
        :warn,
        "Error updating runner #{id} on #{host}."
      )

      false
    end

    def runner_update_needed?(host, token, id, options)
      log_chef(
        :info,
        "Checking runner update required on server #{host} for runner #{id}."
      )

      response = api_runner_get(host, token, id)
      unless api_response_is?(response, Net::HTTPSuccess)
        raise_error(
          message: "Error #{response.code} checking runner update required."
        )
      end

      log_chef(:debug, "Runner found, response code: #{response.code}.")

      runner = JSON.parse(response.body)

      unless runner_options_diff(runner, options).empty?
        log_chef(:info, "Runner #{id} requires update.")
        return true
      end

      log_chef(:info, "Runner #{id} does not require update.")

      false
    end

    def runner_options_diff(server, local)
      log_chef(:debug, "Server runner: #{server}")
      log_chef(:debug, "Local options: #{local}")

      local.reject do |option, value|
        log_chef(
          :info,
          "Testing option #{option}. Local: #{value}. "\
          "Remote: #{server[option]}."
        )

        unless safe_compare(server[option], value)
          log_chef(
            :info,
            "Option #{option} differs. "\
            "Local: #{value}. Server: #{server[option]}."
          )
        end
        safe_compare(server[option], value)
      end
    end

    def runner_get_token(name, config_file)
      toml_load(config_file)['runners']&.each do |r|
        if runner_split_id_name(r['name']).last.eql?(name)
          log_chef(
            :info,
            "Found runner #{name} with token #{r['token']} "\
            "registered at url #{r['url']}."
          )
          return [r['token'], r['url']]
        end
      end

      log_chef(:debug, "Runner #{name} not found.")

      nil
    end

    def runner_get_index(runners, url, token)
      i = runners.find_index do |r|
        r['url'].eql?(url) && r['token'].eql?(token)
      end

      if i.nil?
        raise_error(
          message: "Unable to find index for token #{token} at url #{url}."
        )
      end

      log_chef(:info, "Found runner configuration at index #{i}.")

      i
    end

    def runner_get_id(runner_name, config_file)
      toml_load(config_file)['runners']&.each do |r|
        id, name = runner_split_id_name(r['name'])

        if name.eql?(runner_name)
          log_chef(:info, "Found ID #{id} for runner #{runner_name}.")
          return id
        end
      end

      log_chef(:warn, "No ID found for runner #{runner_name}.")

      nil
    end

    def compact_hash(hash)
      p = proc do |*args|
        v = args.last
        v.delete_if(&p) if v.respond_to?(:delete_if)
        v.nil? || v.respond_to?(:empty?) && v.empty?
      end
      hash.delete_if(&p)
    end

    private

    # Remove options that have been deleted
    def option_cleanup(config:, options:)
      config.delete_if do |option, _|
        next if %w(name token).include?(option)

        unless options.key?(option)
          log_chef(
            :debug,
            "Deleting removed option #{option}."
          )
        end

        !options.key?(option) || RUNNER_LOCAL_REMOVE.include?(option)
      end
    end

    def option_join(options)
      return options if options.empty?

      options.map do |option, value|
        [option, value.respond_to?(:join) ? value.join(',') : value]
      end.to_h
    end

    def runner?(option)
      option.match(RUNNER_REGEXP)
    end

    def runner_option_split(option)
      option.split(RUNNER_REGEXP).delete_if do |string|
        string.empty? || %w(- _).include?(string)
      end
    end

    def format_keyname(key, prefix)
      keyname = key.to_s.tr('_', '-')
      keyname.prepend("#{prefix}-") unless prefix.nil?
      keyname.downcase
    end

    def runner_split_id_name(name)
      return [nil, name] if name.split(':').count.eql?(1)

      result = [name.split(':').first.to_i, name.split(':')[1].strip]
      log_chef(:debug, "Split result: #{result}.")

      result
    end

    def safe_compare(value1, value2)
      v1 = value1.is_a?(Array) ? value1.sort : value1
      v2 = value2.is_a?(Array) ? value2.sort : value2

      v1.eql?(v2)
    end
  end
end
